/*--------------------------------------------------------------------
 |		Copyright 1998 ( ralao@venus.org ) 
 +--------------------------------------------------------------------
 |
 |      WDM RSA Network source file 
 |      by ralao ( ENST 1998 )
 |
 |		This file may be copied, modified and freely distributed as
 |		long as this notice is provided.
 |
 |		Disclaimer :
 |      ------------
 |			This file is part of a Network simulator program. It is 
 |		provided AS IS. Though it has been tested and found harmless 
 |		for any computer, venus.org may not be held liable for any 
 |		damage resulting in the usage of this file.
 |
 +---------------------------------------------------------------------
 |		For assistance, you may contact ralao@venus.org
 +--------------------------------------------------------------------*/



/*--------------------------------------------------------------------
 |	Include definitions bellow this line
 +--------------------------------------------------------------------*/
#include "wdm_rsa_network.h"
#include "wdm_csa_station.h"
#include "assert.h"
#include "malloc.h"
#include "stdlib.h"

/*--------------------------------------------------------------------
 |	Function bodies bellow this line
 +--------------------------------------------------------------------*/

/*--------------------------------------------------------------------
 |	CWDM_RSA_Network :  public constructor
 +--------------------------------------------------------------------*/
CWDM_RSA_Network::CWDM_RSA_Network( long N_station, long k_network, double  p_arr ) : 
    CWDM_Network( N_station, k_network, p_arr )
{
    int i = 0;

    /* creates the N CSA_stations of the network */
    WS  = ( CWDM_Station ** )malloc( N_station * sizeof( CWDM_CSA_Station * ) );
    assert( WS );
    for ( i = 0; i < N_station; i++ ) {
        WS[i]   = new CWDM_CSA_Station( i, N_station, k_network, p_arr );
        assert( WS );
    }
}



/*--------------------------------------------------------------------
 |	~CWDM_RSA_Network : public virtual destructor
 +--------------------------------------------------------------------*/
CWDM_RSA_Network::~CWDM_RSA_Network( void )
{
    int i = 0;
    
    /* deletes the N CSA station created above */
    for ( i = 0; i < N; i++ ) {
        delete WS[i];
    }
    free( WS );
    // C++ calls base class default destructor automatically after 
}


/*----------------------------------------------------------------------------------
 |	simulate :  simulates the activity of the network for a simu_length slot period
 +----------------------------------------------------------------------------------*/
void CWDM_RSA_Network::simulate( long simu_length )
{
    int     i = 0;
    long    old_date    = this->date;
    long    p_delay     = 0;    // propagation delay of CS
    WDM_Slot    dummy = { false, 0, 0, 0 };

    for ( ; this->date < old_date + simu_length; this->date ++ ) {
        /* generates the packets  */
        for ( i = 0; i < N; i++ ) {
            WS[i]->generate_packet( this->date, SP.npg, SP.npb, SP.npl );
        }
        // Caution : Simulation Probe
        for ( i = 0; i < N*N; i ++ ) {
            SP.nsr  += ( CS[0][i] == 1 ) ? 1 : 0;
        }
        /* runs the RSA algorithm */
        { // RSA algorithm 
			bool	done	= false;	// indicates if the algo is done or not
			long	npt		= 0;		// number of packet transmited	

			/* checks if there is something to transmit */
			i = 0;
			while ( i < N*N && CS[0][i] != 1 ) i++;
			if ( i == N*N ) done = true;
			/* starts the algo if necessary */
            while ( !done ) {
				bool	found	= false;	
				long	t_rand	= rand() % N;	// transmitter random number
				long	r_rand	= rand() % N;	// receiver random number  
				long	c_trans	= 0;			// current transmitter being analysed
				long	c_recvr	= 0;			// current receiver being analysed
				long	nvt		= -1;			// number of valid transitter
				long	nvr		= -1;			// number of valid receiver 
								
				/* chooses the right transmitter among the valid and unselected ones 
				and according to t_rand */
				while ( !found ) {
					i	= 0;
					/* checks if c_trans is valid or not */
					while ( i < N && CS[0][ c_trans*N + i] != 1 ) i++;
					/* if so inc the nvt */
					if ( i < N ) nvt++;
					/* if nvt == t_rand then we found one */
					found = ( nvt == t_rand ) ? true : false;
					/* inc c_trans mod N if !found */
					if ( !found ) c_trans = ( c_trans + 1 ) % N;
				}
				/* chooses the right receiver among the valid and unselected ones 
				and according to r_rand */
				while ( nvr != r_rand ) {
					/* checks if c_recvr is valid or not */
					if ( CS[0][ c_trans * N + c_recvr ] == 1 ) nvr++;
					/* if nvr != r_rand inc c_recvr mod N */
					if ( nvr != r_rand ) c_recvr = ( c_recvr + 1 ) % N;
				}
                /* sends the first packet of TB[c_recvr] of WS[c_trans] */
                DS[c_trans]  = WS[c_trans]->send_packet( c_recvr, 0 );
                assert( DS[c_trans].occupied );
                // Caution : Simulation Probe
				npt++;
                SP.tsd += ( this->date - DS[c_trans].arrival_date );
				/* invalidates virtually TB[c_recvr] in all WS */
                for ( i = 0; i < N; i++ ) {
                    CS[0][ i * N + c_recvr] = -1;
                }
				/* invalidates virtually all TBs in WS[c_trans] */
				for ( i = 0; i < N; i++ ) {
                    CS[0][ c_trans * N + i ] = -1;
                }
				/* checks if there is something to transmit anymore */
				i = 0;
				while ( i < N*N && CS[0][i] != 1 ) i++;
				if ( i == N*N ) done = true;
			}
			SP.npt	+= npt;
			SP.nra	+= npt;
			SP.nes	+= ( N - npt );
        }
        /* generates the control slot if the old one has been propagated */
        if ( !p_delay ) {
            for ( i = 0; i < N; i++ ) {        
                WS[i]->update_control_slot( CS[k] );
            }
        }
        p_delay = ( p_delay + 1 ) % this->k ;
        /* propagates the control slot for one slot*/
        this->propagate_control_slot();
        /* updates the simulation probe */
        SP.nsg  += N;
        SP.q    = ( SP.nsr ) ?  1.0 * SP.nra / SP.nsr   : SP.q;
        SP.ro   = ( SP.nsg ) ?  1.0 * SP.npg / SP.nsg   : SP.ro;
        SP.asd  = ( SP.npt ) ?  1.0 * SP.tsd / SP.npt   : SP.asd;
        SP.apl  = ( SP.npg ) ?  1.0 * SP.npl / SP.npg   : SP.apl;
        SP.ant  = ( SP.nsg ) ?  1.0 * SP.npt / SP.nsg   : SP.ant;
    }
}
